//
//  iconc.cpp (http://exoton.com/free/iconc)
//
//  Console utility to compile Windows .ico file from one or more images.
//
//  Images can be in any format supported by GDI+ (BMP, GIF, JPEG, PNG, TIFF...).
//  Images in icon file are in the PixelFormat32bppARGB format.
//  Images bigger than 128x128 are copied as is (PNG should work).
//
//  Copyright 2014 (c) Exoton Inc.
//
//  Licensed under MIT (http://exoton.com/free/LICENSE)
//______________________________________________________________________________

#include <windows.h>
#include <gdiplus.h>
#include <stdio.h>
#include <sys/stat.h>

using namespace Gdiplus;

//
//  http://msdn.microsoft.com/en-us/library/ms997538.aspx
//

typedef struct {
	BYTE  bWidth;        // Width, in pixels, of the image
	BYTE  bHeight;       // Height, in pixels, of the image
	BYTE  bColorCount;   // Number of colors in image (0 if >=8bpp)
	BYTE  bReserved;     // Reserved (must be 0)
	WORD  wPlanes;       // Color Planes
	WORD  wBitCount;     // Bits per pixel
	DWORD dwBytesInRes;  // How many bytes in this resource?
	DWORD dwImageOffset; // Where in the file is this image?
} ICONDIRENTRY;

static UINT ICONDIR_SIZE      = 3 * sizeof(WORD);
static UINT ICONDIRENTRY_SIZE = sizeof(ICONDIRENTRY);

static inline UINT mask_width_bytes(UINT w) { return ((w + 31) / 8) & ~3; } // multiple of 4
//_______________.._______________


static UINT imageSize(UINT w, UINT h)
{
	return (w * 4 + mask_width_bytes(w)) * h;
}
//_______________.._______________

static void writeIconDir(HANDLE hFile, int count)
{
	DWORD nWritten;
	WORD  icondir[] = {0, 1, count};

	WriteFile(hFile, &icondir, ICONDIR_SIZE, &nWritten, 0);
}
//_______________.._______________

static void writeIconDirEntry(HANDLE hFile, UINT w, UINT h, UINT size, UINT off)
{
	ICONDIRENTRY ide;
	ide.bWidth = w;
	ide.bHeight = h;
	ide.bColorCount = 0;
	ide.bReserved = 0;
	ide.wPlanes = 1;
	ide.wBitCount = 32;
	ide.dwBytesInRes = size;
	ide.dwImageOffset = off;

	DWORD nWritten;
	WriteFile(hFile, &ide, ICONDIRENTRY_SIZE, &nWritten, 0);
}
//_______________.._______________

static void addFile(HANDLE hFile, const wchar_t* path, UINT w, UINT h)
{
	struct _stat st;
	_wstat(path, &st);

	void* p = malloc(st.st_size);
	FILE* fp = _wfopen(path, L"rb");
	fread(p, 1, st.st_size, fp);
	fclose(fp);
	DWORD nWritten;
	WriteFile(hFile, p, st.st_size, &nWritten, 0);
	free(p);
}
//_______________.._______________

static void addIcon(HANDLE hFile, Bitmap* bitmap)
{
	UINT w = bitmap->GetWidth();
	UINT h = bitmap->GetHeight();

	Rect rect(0, 0, w, h);
	BitmapData data;
	bitmap->LockBits(&rect, ImageLockModeRead, PixelFormat32bppARGB, &data);

	char* p = (char*)data.Scan0;
	int stride = data.Stride;
	if (stride > 0)
		p += (h - 1) * stride;

	BITMAPINFOHEADER bih;
	ZeroMemory(&bih, sizeof(bih));
	bih.biSize = sizeof(bih);
	bih.biWidth = (LONG)w;
	bih.biHeight = (LONG)h * 2;
	bih.biPlanes = 1;
	bih.biBitCount = 32;
	bih.biCompression = BI_RGB;
	bih.biSizeImage = imageSize(w, h);

	DWORD nWritten;
	WriteFile(hFile, &bih, sizeof(bih), &nWritten, 0);

	for (UINT i = 0; i < h; ++i, p -= stride)
		WriteFile(hFile, p, w * 4, &nWritten, 0);

	UINT n = mask_width_bytes(w) * h;
	memset(data.Scan0, 0, n);
	WriteFile(hFile, data.Scan0, n, &nWritten, 0);

	bitmap->UnlockBits(&data);
}
//_______________.._______________

int wmain(int argc, const wchar_t* argv[])
{
	if (argc < 3) {
		wprintf(L"Usage: %s input [input ...] output\n", argv[0]);
		return 1;
	}

	ULONG_PTR gdiplusToken;
	GdiplusStartupInput gdiplusStartupInput;
	if (GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL)) {
		wprintf(L"Failed to initializes Windows GDI+\n");
		return 1;
	}

	HANDLE hFile = INVALID_HANDLE_VALUE;
	int r = 1;
	int n = argc - 2;
	Bitmap** bitmaps = (Bitmap**)calloc(n, sizeof(Bitmap*));

	for (int i = 0; i < n; ++i) {
		bitmaps[i] = new Bitmap(argv[i + 1]);
		if (bitmaps[i]->GetType() != ImageTypeBitmap) {
			wprintf(L"Failed to load image file '%s'\n", argv[i + 1]);
			goto exit;
		}
	}

	hFile = CreateFile(argv[argc - 1], GENERIC_WRITE, 0, 0, CREATE_ALWAYS, 0, 0);
	if (hFile == INVALID_HANDLE_VALUE) {
		wprintf(L"Failed to create icon file '%s'\n", argv[argc - 1]);
		goto exit;
	}

	writeIconDir(hFile, n);
	UINT off = ICONDIR_SIZE + ICONDIRENTRY_SIZE * n;

	for (int i = 0; i < n; ++i) {
		UINT size;
		UINT w = bitmaps[i]->GetWidth();
		UINT h = bitmaps[i]->GetHeight();
		if (w > 255 || h > 255)
			w = h = 0;
		if (w  == 0 || w * h > 128 * 128) {
			struct _stat st;
			_wstat(argv[i + 1], &st);
			size = st.st_size;
		} else {
			size = sizeof(BITMAPINFOHEADER) + imageSize(w, h);
		}
		writeIconDirEntry(hFile, w, h, size, off);
		off += size;
	}
	for (int i = 0; i < n; ++i) {
		UINT w = bitmaps[i]->GetWidth();
		UINT h = bitmaps[i]->GetHeight();
		if (w > 255 || h > 255 || w * h > 128 * 128)
			addFile(hFile, argv[i + 1], w, h);
		else
			addIcon(hFile, bitmaps[i]);
	}
	r = 0;

exit:
	if (hFile != INVALID_HANDLE_VALUE)
		CloseHandle(hFile);
	for (int i = 0; i < n; ++i)
		delete bitmaps[i];
	GdiplusShutdown(gdiplusToken);

	return r;
}
